import numpy as np

from ..mutual_information import conditional_mutual_information


def iamb(X, lamb=0.0, method=None, options=None):
    """Search Markov blanket in a Bayesian network using
       Incremental Association Markov Blanket algorithm [1]_.
    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence test.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    mb : array, shape (d, d)
        Estimated Markov blanket.
    References
    ----------
    .. [1] Tsamardinos, Ioannis, et al. "Algorithms for
           Large Scale Markov Blanket Discovery." FLAIRS
           conference. Vol. 2. 2003.
    """
    n, d = X.shape
    mb = np.zeros([d, d], dtype=bool)
    while True:
        cmi_max = -np.inf
        for i in range(d):
            x = X[:, i]
            z = X[:, mb[i]]
            non_mb = ~mb[i] & (np.arange(d) != i)
            for j in non_mb.nonzero()[0]:
                y = X[:, j]
                cmi = conditional_mutual_information(x, y, z, method, options)
                if cmi_max < cmi:
                    cmi_max = cmi
                    idx_max = i, j

        if cmi_max <= lamb:
            break

        mb[idx_max] = 1

    for i in range(d):
        x = X[:, i]
        for j in mb[i].nonzero()[0]:
            other_mb = mb[i] & (np.arange(d) != j)
            y = X[:, j]
            z = X[:, other_mb]
            cmi = conditional_mutual_information(x, y, z, method, options)
            if cmi <= lamb:
                mb[i, j] = 0

    return mb


def mmpc(X, lamb=0.0, method=None, options=None):
    """Search parents and children in a Bayesian network.
    Parameters
    ----------
    X : array, shape (n_samples, d)
        Observations of variables.
    lamb: float
        Threshold for independence test.
    method: str, default 'knn'
        Method for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    ----------
    pc : array, shape (d, d)
        Estimated parents and children.
    """
    n, d = X.shape
    pc = np.zeros([d, d], dtype=bool)
    for i in range(d):
        x = X[:, i]
        non_pc = ~pc[i] & (np.arange(d) != i)
        for j in non_pc.nonzero()[0]:
            y = X[:, j]
            z = X[:, pc[i]]
            cmi = conditional_mutual_information(x, y, z, method=method, options=options)
            if cmi > lamb:
                pc[i, j] = 1

        for j in pc[i].nonzero()[0]:
            other_pc = pc[i] & (np.arange(d) != j)
            y = X[:, j]
            z = X[:, other_pc]
            cmi = conditional_mutual_information(x, y, z, method=method, options=options)
            if cmi <= lamb:
                pc[i, j] = 0

    return pc & pc.T



import numpy as np
from scipy.special import digamma


def _mi_knn(X, Y, k):
    n, d_x = X.shape
    _, d_y = Y.shape
    distances_x = np.linalg.norm(X - X.reshape([n, -1, d_x]), axis=2)
    distances_y = np.linalg.norm(Y - Y.reshape([n, -1, d_y]), axis=2)
    distances = np.maximum(distances_x, distances_y)
    epsilons = np.partition(distances, k, axis=1)[:, k]
    idx_discrete = np.isclose(epsilons, 0)
    ks = np.repeat(k, n)
    ks[idx_discrete] = np.sum(np.isclose(distances[idx_discrete], 0), axis=1) - 1
    n_x = np.sum(distances_x <= epsilons, axis=0) - 1
    n_y = np.sum(distances_y <= epsilons, axis=0) - 1
    mi = np.log(n) + np.mean(digamma(ks) - np.log(n_x * n_y))
    return mi



import numpy as np
from ..mutual_information import mutual_information


def _mrmr(X, y, lamb=0.0, method='knn', options=None):
    """Select effective features in X on predinting y using
       minimum redundancy maximum relevance feature selection [1]_.
    Parameters
    ----------
    X : array_like, shape (n_samples, d)
        Features.
    y : array_like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float
        Threshold for independence tests.
    method: str, default 'knn'
        Method used for MI estimation.
    options : dict, default None
        Optional parameters for MI estimation.
    Returns
    -------
    indices : list
        Indices for the selected features.
    References
    ----------
    .. [1] Peng, Hanchuan, Fuhui Long, and Chris Ding. "Feature
           selection based on mutual information criteria of
           max-dependency, max-relevance, and min-redundancy."
           IEEE Transactions on pattern analysis and machine
           intelligence 27.8 (2005): 1226-1238.
    """
    n, d = X.shape
    selected = []
    while True:
        max_obj = -np.inf
        not_selected = set(range(d)) - set(selected)
        for i in not_selected:
            rel = mutual_information(X[:, i], y)
            red = [mutual_information(X[:, i], X[:, j]) for j in selected]
            obj = rel - (np.mean(red) if red else 0)
            if max_obj < obj:
                max_obj = obj
                max_idx = i

        if max_obj <= 0:
            return selected

        selected.append(max_idx)



from .mrmr_ import _mrmr
from .mifs_ import _mifs


def select(X, y, lamb=0.0, method='mifs'):
    """Select effective features in X on predinting y using
    Parameters
    ----------
    X : array_like, shape (n_samples, d)
        Features.
    y : array_like, shape (n_samples)
        Target. Can be either continuous or discrete.
    lamb: float, default 0.0
        Threshold for independence tests.
    method: str, default 'mifs'
        Method used for feature selection. Either 'mifs' or
        'mrmr' can be chosen.
    Returns
    -------
    indices : list
        Indices for the selected features.
    """
    if method == 'mifs':
        return _mifs(X, y, lamb=0.0, method='knn', options=None)
    elif method == 'mrmr':
        return _mrmr(X, y, lamb=0.0, method='knn', options=None)
    else:
        raise NotImplementedError()
